//
//  Copyright © 2017 Jan Gorman. All rights reserved.
//

import XCTest
import UIKit
import MapleBacon
#if canImport(Combine)
import Combine
#endif

final class MapleBaconCacheTests: XCTestCase {
  
  private let helper = TestHelper()

  @available(iOS 13.0, *)
  private lazy var subscriptions: Set<AnyCancellable> = []

  override func tearDown() {
    super.tearDown()
    if #available(iOS 13.0, *) {
      subscriptions.removeAll()
    }
  }
  
  func testItStoresImageInMemory() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let key = "http://\(#function)"

    let expectation = self.expectation(description: #function)
    
    cache.store(data: imageData, forKey: key) {
      cache.retrieveImage(forKey: key) { image, type in
        XCTAssertNotNil(image)
        XCTAssertEqual(type, .memory)
        expectation.fulfill()
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testNamedCachesAreDistinct() {
    let mockCache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let namedCache = MapleBaconCache(name: "named")
    let imageData = helper.imageData
    let key = #function

    let expectation = self.expectation(description: #function)

    mockCache.store(data: imageData, forKey: key) {
      namedCache.retrieveImage(forKey: key) { image, _ in
        XCTAssertNil(image)
        expectation.fulfill()
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testUnknownCacheKeyReturnsNoImage() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData

    let expectation = self.expectation(description: #function)

    cache.store(data: imageData, forKey: "key1") {
      cache.retrieveImage(forKey: "key2") { image, type in
        XCTAssertNil(image)
        XCTAssertEqual(type, .none)
        expectation.fulfill()
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }
  
  func testItStoresImagesToDisk() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let key = #function

    let expectation = self.expectation(description: #function)

    cache.store(data: imageData, forKey: key) {
      cache.clearMemory()
      cache.retrieveImage(forKey: key) { image, type in
        XCTAssertNotNil(image)
        XCTAssertEqual(type, .disk)
        expectation.fulfill()
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testImagesOnDiskAreMovedToMemory() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let key = #function

    let expectation = self.expectation(description: #function)

    cache.store(data: imageData, forKey: key) {
      cache.clearMemory()
      cache.retrieveImage(forKey: key) { _, _ in
        cache.retrieveImage(forKey: key) { image, type in
          XCTAssertNotNil(image)
          XCTAssertEqual(type, .memory)
          expectation.fulfill()
        }
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testItClearsDiskCache() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let key = #function

    let expectation = self.expectation(description: #function)

    cache.store(data: imageData, forKey: key) {
      cache.clearMemory()
      cache.clearDisk {
        cache.retrieveImage(forKey: key) { image, _ in
          XCTAssertNil(image)
          expectation.fulfill()
        }
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testCacheWithIdentifierIsCachedAsSeparateImage() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let alternateImageData = helper.image.jpegData(compressionQuality: 0.2)!
    let key = #function
    let transformerId = "transformer"

    let expectation = self.expectation(description: #function)

    cache.store(data: imageData, forKey: key) {
      cache.store(data: alternateImageData, forKey: key, transformerId: transformerId) {
        cache.retrieveImage(forKey: key) { image, _ in
          XCTAssertNotNil(image)

          cache.retrieveImage(forKey: key, transformerId: transformerId) { transformerImage, _ in
            XCTAssertNotNil(transformerImage)
            XCTAssertNotEqual(image, transformerImage)
            expectation.fulfill()
          }
        }
      }
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

}

#if canImport(Combine)

@available(iOS 13.0, *)
extension MapleBaconCacheTests {

  func testItStoresImagesInMemoryPublisher() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let key = "http://\(#function)"

    let expectation = self.expectation(description: #function)

    cache.storeAndPublish(data: imageData, forKey: key)
      .sink { _ in
        cache.retrieveImage(forKey: key)
          .sink { image, type in
            XCTAssertNotNil(image)
            XCTAssertEqual(type, .memory)
            expectation.fulfill()
          }
          .store(in: &self.subscriptions)
      }
      .store(in: &self.subscriptions)

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testItStoresDataInMemoryPublisher() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData
    let key = "http://\(#function)"

    let expectation = self.expectation(description: #function)

    cache.storeAndPublish(data: imageData, forKey: key)
      .sink { _ in
        cache.retrieveData(forKey: key)
          .sink { data, type in
            XCTAssertNotNil(data)
            XCTAssertEqual(type, .memory)
            expectation.fulfill()
          }
          .store(in: &self.subscriptions)
      }
      .store(in: &self.subscriptions)

    waitForExpectations(timeout: 5, handler: nil)
  }

  func testUnknownCacheKeyReturnsNoImagePublisher() {
    let cache = MapleBaconCache(name: "mock", backingStore: MockStore())
    let imageData = helper.imageData

    let expectation = self.expectation(description: #function)

    cache.store(data: imageData, forKey: "key1") {
      cache.retrieveImage(forKey: "key2")
        .sink { image, type in
          XCTAssertNil(image)
          XCTAssertEqual(type, .none)
          expectation.fulfill()
        }
        .store(in: &self.subscriptions)
    }

    waitForExpectations(timeout: 5, handler: nil)
  }

}

#endif
